import isPlainObject from 'lodash/isPlainObject';
import { default as d } from 'debug';

export const tryAll = promiseConstructors => {
  const promise = new Promise(promiseConstructors[0]);
  if (promiseConstructors.length === 1) {
    return promise;
  }
  return promise.catch(() => tryAll(promiseConstructors.slice(1)));
};

export const tap = interceptor => value => (interceptor(value), value);

const debug = d('LC:Cache');
export class Cache {
  constructor(name = 'anonymous') {
    this.name = name;
    this._map = {};
  }

  get(key) {
    const cache = this._map[key];
    if (cache) {
      const expired = cache.expiredAt && cache.expiredAt < Date.now();
      if (!expired) {
        debug(`[${this.name}] hit: ${key} ${cache.value}`);
        return cache.value;
      }
      debug(`[${this.name}] expired: ${key}`);
    }
    debug(`[${this.name}] missed: ${key}`);
    return null;
  }

  set(key, value, ttl) {
    debug(`[${this.name}] set: ${key} ${value} ${ttl}`);
    const cache = this._map[key] = {
      value,
    };
    if (typeof ttl === 'number') {
      cache.expiredAt = Date.now() + ttl;
    }
  }
}

export const decodeDate = date => {
  if (!date) return date;
  if (typeof date === 'string') {
    return new Date(date);
  }
  if (date.__type === 'Date' && date.iso) {
    return new Date(date.iso);
  }
  return date;
};

export const keyRemap = (keymap, obj) =>
  Object.keys(obj).reduce((newObj, key) => {
    const newKey = keymap[key] || key;
    return Object.assign(newObj, {
      [newKey]: obj[key],
    });
  }, {});

export const union = (a, b) => Array.from(new Set([...a, ...b]));
export const difference = (a, b) => Array.from(
  (bSet => new Set(a.filter(x => !bSet.has(x))))(new Set(b))
);

// debug utility
const removeNull = obj => {
  if (!isPlainObject(obj)) return obj;
  const object = Object.assign({}, obj);
  // eslint-disable-next-line no-restricted-syntax
  for (const prop in object) {
    if (object.hasOwnProperty(prop)) {
      const value = object[prop];
      if (value === null) {
        delete object[prop];
      } else {
        object[prop] = removeNull(value);
      }
    }
  }
  return object;
};
export const trim = message => removeNull(JSON.parse(JSON.stringify(message)));
